@@ -1,9 +1,12 @@
 # -*- coding: utf-8 -*-
 #
 # Picard, the next-generation MusicBrainz tagger
+#
 # Copyright (C) 2007 Lukáš Lalinský
 # Copyright (C) 2009 Carlin Mangar
 # Copyright (C) 2017 Sambhav Kothari
+# Copyright (C) 2018-2020 Philipp Wolfer
+# Copyright (C) 2018-2020 Laurent Monin
 #
 # This program is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License
@@ -39,7 +42,6 @@ from PyQt5 import (
     QtNetwork,
 )
 from PyQt5.QtCore import (
-    QStandardPaths,
     QUrl,
     QUrlQuery,
 )
@@ -52,9 +54,14 @@ from picard import (
     config,
     log,
 )
+from picard.const import (
+    CACHE_DIR,
+    CACHE_SIZE_IN_BYTES,
+)
 from picard.oauth import OAuthManager
 from picard.util import (
     build_qurl,
+    bytes2human,
     parse_json,
 )
 from picard.util.xml import parse_xml
@@ -259,9 +266,12 @@ class WebService(QtCore.QObject):
     def __init__(self, parent=None):
         super().__init__(parent)
         self.manager = QtNetwork.QNetworkAccessManager()
+        self._network_accessible_changed(self.manager.networkAccessible())
+        self.manager.networkAccessibleChanged.connect(self._network_accessible_changed)
         self.oauth_manager = OAuthManager(self)
         self.set_cache()
         self.setup_proxy()
+        self.set_transfer_timeout(config.setting['network_transfer_timeout_seconds'])
         self.manager.finished.connect(self._process_reply)
         self._request_methods = {
             "GET": self.manager.get,
@@ -272,6 +282,15 @@ class WebService(QtCore.QObject):
         self._init_queues()
         self._init_timers()
 
+    def _network_accessible_changed(self, accessible):
+        # Qt's network accessibility check sometimes fails, e.g. with VPNs on Windows.
+        # If the accessibility is reported to be not accessible, set it to
+        # unknown instead. Let's just try any request and handle the error.
+        # See https://tickets.metabrainz.org/browse/PICARD-1791
+        if accessible == QtNetwork.QNetworkAccessManager.NotAccessible:
+            self.manager.setNetworkAccessible(QtNetwork.QNetworkAccessManager.UnknownAccessibility)
+        log.debug("Network accessible requested: %s, actual: %s", accessible, self.manager.networkAccessible())
+
     def _init_queues(self):
         self._active_requests = {}
         self._queues = defaultdict(lambda: defaultdict(deque))
@@ -286,26 +305,41 @@ class WebService(QtCore.QObject):
         self._timer_count_pending_requests.setSingleShot(True)
         self._timer_count_pending_requests.timeout.connect(self._count_pending_requests)
 
-    def set_cache(self, cache_size_in_mb=100):
+    def set_cache(self, cache_size_in_bytes=None):
+        if cache_size_in_bytes is None:
+            cache_size_in_bytes = CACHE_SIZE_IN_BYTES
         cache = QtNetwork.QNetworkDiskCache()
-        location = QStandardPaths.writableLocation(QStandardPaths.CacheLocation)
-        cache.setCacheDirectory(os.path.join(location, 'picard'))
-        cache.setMaximumCacheSize(cache_size_in_mb * 1024 * 1024)
+        cache.setCacheDirectory(os.path.join(CACHE_DIR, 'network'))
+        cache.setMaximumCacheSize(cache_size_in_bytes)
         self.manager.setCache(cache)
-        log.debug("NetworkDiskCache dir: %r size: %s / %s",
-                  cache.cacheDirectory(), cache.cacheSize(),
-                  cache.maximumCacheSize())
+        log.debug("NetworkDiskCache dir: %r current size: %s max size: %s",
+                  cache.cacheDirectory(),
+                  bytes2human.decimal(cache.cacheSize(), l10n=False),
+                  bytes2human.decimal(cache.maximumCacheSize(), l10n=False))
 
     def setup_proxy(self):
         proxy = QtNetwork.QNetworkProxy()
         if config.setting["use_proxy"]:
-            proxy.setType(QtNetwork.QNetworkProxy.HttpProxy)
+            if config.setting["proxy_type"] == 'socks':
+                proxy.setType(QtNetwork.QNetworkProxy.Socks5Proxy)
+            else:
+                proxy.setType(QtNetwork.QNetworkProxy.HttpProxy)
             proxy.setHostName(config.setting["proxy_server_host"])
             proxy.setPort(config.setting["proxy_server_port"])
-            proxy.setUser(config.setting["proxy_username"])
-            proxy.setPassword(config.setting["proxy_password"])
+            if config.setting["proxy_username"]:
+                proxy.setUser(config.setting["proxy_username"])
+            if config.setting["proxy_password"]:
+                proxy.setPassword(config.setting["proxy_password"])
         self.manager.setProxy(proxy)
 
+    def set_transfer_timeout(self, timeout):
+        timeout_ms = timeout * 1000
+        if hasattr(self.manager, 'setTransferTimeout'):  # Available since Qt 5.15
+            self.manager.setTransferTimeout(timeout_ms)
+            self._transfer_timeout = 0
+        else:  # Use fallback implementation
+            self._transfer_timeout = timeout_ms
+
     def _send_request(self, request, access_token=None):
         hostkey = request.get_host_key()
         ratecontrol.increment_requests(hostkey)
@@ -314,8 +348,32 @@ class WebService(QtCore.QObject):
         send = self._request_methods[request.method]
         data = request.data
         reply = send(request, data.encode('utf-8')) if data is not None else send(request)
+        self._start_transfer_timeout(reply)
         self._active_requests[reply] = request
 
+    def _start_transfer_timeout(self, reply):
+        if not self._transfer_timeout:
+            return
+        # Fallback implementation of a transfer timeout for Qt < 5.15.
+        # Aborts a request if no data gets transferred for TRANSFER_TIMEOUT milliseconds.
+        timer = QtCore.QTimer(self)
+        timer.setSingleShot(True)
+        timer.setTimerType(QtCore.Qt.PreciseTimer)
+        timer.timeout.connect(partial(self._timeout_request, reply))
+        reply.finished.connect(timer.stop)
+        reset_callback = partial(self._reset_transfer_timeout, timer)
+        reply.uploadProgress.connect(reset_callback)
+        reply.downloadProgress.connect(reset_callback)
+        timer.start(self._transfer_timeout)
+
+    def _reset_transfer_timeout(self, timer, bytesTransferred, bytesTotal):
+        timer.start(self._transfer_timeout)
+
+    @staticmethod
+    def _timeout_request(reply):
+        if reply.isRunning():
+            reply.abort()
+
     def _start_request(self, request):
         if request.mblogin and request.path != "/oauth2/token":
             self.oauth_manager.get_access_token(partial(self._send_request, request))
